==> atmos-2.96.9.tar.gz: atmos-2.96.9/src/spectral_factory.cpp <==
/*
    $Id: spectral_factory.cpp 158 2010-09-19 08:38:25Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iomanip>
#include <fstream>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem/operations.hpp>

#include <boost/numeric/ublas/io.hpp>

#include "weif.h"
#include "utils.h"

#include "spectral_factory.h"

using namespace boost::numeric::ublas;

std::pair< spectral::vector_type, spectral::vector_type > spectral::load_reaction( const boost::filesystem::path& file ) {
	std::ifstream fstm( file.string().c_str() );
	if( fstm.fail() )
		throw std::runtime_error( "Can't open file " + file.string() );

	std::pair< vector_type, vector_type > ret;
	vector< double >::size_type alloc_size = 10;
	vector< double >::size_type size = 0;

	ret.first.resize( alloc_size );
	ret.second.resize( alloc_size );

	std::string cur_line;
	double cx,cy;
	while( std::getline( fstm, cur_line ) ) {
		boost::trim_if( cur_line, boost::is_space() );
		if( cur_line.size() < 1 || cur_line.at(0) == '#' ) continue;
		std::istringstream cstm( cur_line );
		if( !(cstm >> cx >> cy) )
			throw std::runtime_error( "File format error in " + file.string() );
		ret.first( size ) = cx;
		ret.second( size ) = cy;
		++size;
		if( size >= ret.first.size() ) {
			alloc_size += alloc_size;
			ret.first.resize( alloc_size );
			ret.second.resize( alloc_size );
		}
	}

	ret.first.resize( size );
	ret.second.resize( size );

	/* Post check */
	if( size == 0 )
		throw std::runtime_error( "File " + file.string() + " contains no data" );
	if( std::adjacent_find( ret.first.begin(), ret.first.end(), std::greater<double>() ) != ret.first.end() )
		throw std::runtime_error( "File " + file.string() + " contains unsorted data" );
	if( std::count_if( ret.first.begin(), ret.first.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "File " + file.string() + " contains negative wavelengths" );
	if( std::count_if( ret.second.begin(), ret.second.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "File " + file.string() + " contains negative data" );

	ret.first = ret.first * 0.001; // convert from nm to mkm

	return ret;
}

std::pair< spectral::vector_type, spectral::vector_type > spectral::photon_distribution( const vector_type& spectral_lambda, const vector_type& spectral_response, const vector_type& device_lambda, const vector_type& device_response ) {
	std::pair< vector_type, vector_type > ret;

	assert( spectral_lambda.size() == spectral_response.size() );
	assert( device_lambda.size() == device_response.size() );

	ret.first.resize( std::min(spectral_lambda.size(), device_lambda.size()) );

	vector_type::iterator it = std::set_intersection( spectral_lambda.begin(), spectral_lambda.end(), device_lambda.begin(), device_lambda.end(), ret.first.begin() );
	if( it - ret.first.begin() == 0 )
		throw std::runtime_error( "No match of response and spectrum grids" );

	ret.first.resize( it - ret.first.begin() );
	ret.second.resize( ret.first.size() );

	vector_type::const_iterator sit = spectral_lambda.begin();
	vector_type::const_iterator dit = device_lambda.begin();
	vector_type::iterator rit = ret.second.begin();
	for( vector_type::const_iterator it = ret.first.begin(); it != ret.first.end(); ++it ) {
		sit = std::find( sit, spectral_lambda.end(), *it );
		dit = std::find( dit, device_lambda.end(), *it );
		*rit = spectral_response( sit.index() ) * device_response( dit.index() );
		++rit;
	}

	if( std::count_if( ret.second.begin(), ret.second.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "SED contains negative data" );

	ret.second = ret.second / sum( ret.second );

	return ret;
}

void spectral::dump_weight( const boost::filesystem::path& file, const boost::function1< vector_type, double >& function, const vector_type& grid ) {
	static const char* name[] = { "A", "B", "C", "D", "AB", "AC", "AD", "BC", "BD", "CD", "L", "T" };
	static const unsigned int name_size = sizeof(name) / sizeof(name[0]);
	const double wscale = 1e+11;

	std::ofstream fstm( file.string().c_str() );
	if( fstm.fail() )
		throw std::runtime_error( "Can't open file " + file.string() );

	fstm.exceptions( std::ifstream::failbit | std::ifstream::badbit );

	fstm << "# Z " << std::setprecision(2) << grid << std::endl;
	fstm << "# W " << std::setprecision(0) << std::scientific << wscale << std::endl;
	if( grid.size() < 1 )
		 return;

	unsigned int result_size = function( grid(0) ).size();

	fstm << "## Alt";
	for( unsigned int i = 0; i < std::min(name_size, result_size); ++i )
		fstm << "\tW(" << name[i] << ")/" <<  std::scientific << wscale;
	fstm << std::endl;

	for( vector_type::size_type i = 0; i < grid.size(); ++i ){
		fstm << std::fixed << std::setprecision(3) << grid(i);
		vector_type result = function( grid(i) );
		fstm << std::scientific << std::setprecision(4);
		for( vector_type::size_type j = 0; j < result.size(); ++j ) {
			fstm << "\t" << result(j) / wscale;
		}
		fstm << std::endl;
	}
}

spectral_factory::spectral_factory(const boost::filesystem::path& wd, const boost::filesystem::path& mass_response, const boost::filesystem::path& ccd_response ):
	m_mass_response(mass_response),
	m_ccd_response(ccd_response),
	m_wdir(wd),
	m_basis_grid( lg_generator(0.0, 0.04, 0.25), 50 ),
	m_mass_reaction( spectral::load_reaction(mass_response) ),
	m_dimm_reaction( spectral::load_reaction(ccd_response) ) {

	if( ! (boost::filesystem::exists( wd ) && boost::filesystem::is_directory( wd )) )
		throw std::runtime_error( wd.string() + " is not a directory or doesn't exists");
}
spectral_factory::~spectral_factory() {

}
const boost::filesystem::path spectral_factory::get_spectrum_filename( const std::string& spectype ) const {
	std::string spectral_type( spectype );
	boost::to_lower( spectral_type );
	boost::filesystem::path spect( m_wdir / (spectral_type + ".sp") );
	return spect;
}
const spectral_factory::weight_type& spectral_factory::spectral( const std::string& spectype, const vector<double>& inner_diam, const vector<double>& outer_diam, double dimmbase, double dimmsize ) {
	storage_type::const_iterator it = m_storage.find( spectype );
	if( it != m_storage.end() )
		return it->second;

	std::pair< vector_type, vector_type > spectral_response = spectral::load_reaction( get_spectrum_filename( spectype ) );

	std::pair< vector_type, vector_type > mass_response = spectral::photon_distribution( spectral_response.first, spectral_response.second, m_mass_reaction.first, m_mass_reaction.second );
	std::pair< vector_type, vector_type > dimm_response = spectral::photon_distribution( spectral_response.first, spectral_response.second, m_dimm_reaction.first, m_dimm_reaction.second );

	std::pair< storage_type::iterator, bool > ret = m_storage.insert( std::make_pair( spectype, weight_type( m_basis_grid, utils::ublas::project( m_basis_grid , weif::massdimm_weight( mass_response.first, mass_response.second, dimm_response.first, dimm_response.second, inner_diam, outer_diam, dimmsize, dimmbase ) ) ) ) );

	return ret.first->second;
}

==> atmos-2.98.1.tar.gz: atmos-2.98.1/src/spectral_factory.cpp <==
/*
    $Id: spectral_factory.cpp 177 2010-12-13 10:37:30Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iomanip>
#include <fstream>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem/operations.hpp>

#include <boost/numeric/ublas/io.hpp>

#include "weif.h"
#include "utils.h"

#include "spectral_factory.h"

using namespace boost::numeric::ublas;

std::pair< spectral::vector_type, spectral::vector_type > spectral::load_reaction( const boost::filesystem::path& file ) {
	std::ifstream fstm( file.string().c_str() );
	if( fstm.fail() )
		throw std::runtime_error( "Can't open file " + file.string() );

	std::pair< vector_type, vector_type > ret;
	vector< double >::size_type alloc_size = 10;
	vector< double >::size_type size = 0;

	ret.first.resize( alloc_size );
	ret.second.resize( alloc_size );

	std::string cur_line;
	double cx,cy;
	while( std::getline( fstm, cur_line ) ) {
		boost::trim_if( cur_line, boost::is_space() );
		if( cur_line.size() < 1 || cur_line.at(0) == '#' ) continue;
		std::istringstream cstm( cur_line );
		if( !(cstm >> cx >> cy) )
			throw std::runtime_error( "File format error in " + file.string() );
		ret.first( size ) = cx;
		ret.second( size ) = cy;
		++size;
		if( size >= ret.first.size() ) {
			alloc_size += alloc_size;
			ret.first.resize( alloc_size );
			ret.second.resize( alloc_size );
		}
	}

	ret.first.resize( size );
	ret.second.resize( size );

	/* Post check */
	if( size == 0 )
		throw std::runtime_error( "File " + file.string() + " contains no data" );
	if( std::adjacent_find( ret.first.begin(), ret.first.end(), std::greater<double>() ) != ret.first.end() )
		throw std::runtime_error( "File " + file.string() + " contains unsorted data" );
	if( std::count_if( ret.first.begin(), ret.first.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "File " + file.string() + " contains negative wavelengths" );
	if( std::count_if( ret.second.begin(), ret.second.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "File " + file.string() + " contains negative data" );

	ret.first = ret.first * 0.001; // convert from nm to mkm

	return ret;
}

std::pair< spectral::vector_type, spectral::vector_type > spectral::photon_distribution( const vector_type& spectral_lambda, const vector_type& spectral_response, const vector_type& device_lambda, const vector_type& device_response ) {
	std::pair< vector_type, vector_type > ret;

	assert( spectral_lambda.size() == spectral_response.size() );
	assert( device_lambda.size() == device_response.size() );

	ret.first.resize( std::min(spectral_lambda.size(), device_lambda.size()) );

	vector_type::iterator it = std::set_intersection( spectral_lambda.begin(), spectral_lambda.end(), device_lambda.begin(), device_lambda.end(), ret.first.begin() );
	if( it - ret.first.begin() == 0 )
		throw std::runtime_error( "No match of response and spectrum grids" );

	ret.first.resize( it - ret.first.begin() );
	ret.second.resize( ret.first.size() );

	vector_type::const_iterator sit = spectral_lambda.begin();
	vector_type::const_iterator dit = device_lambda.begin();
	vector_type::iterator rit = ret.second.begin();
	for( vector_type::const_iterator it = ret.first.begin(); it != ret.first.end(); ++it ) {
		sit = std::find( sit, spectral_lambda.end(), *it );
		dit = std::find( dit, device_lambda.end(), *it );
		*rit = spectral_response( sit.index() ) * device_response( dit.index() );
		++rit;
	}

	if( std::count_if( ret.second.begin(), ret.second.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "SED contains negative data" );

	ret.second = ret.second / sum( ret.second );

	return ret;
}

void spectral::dump_weight( const boost::filesystem::path& file, const boost::function1< vector_type, double >& function, const vector_type& grid ) {
	static const char* name[] = { "A", "B", "C", "D", "AB", "AC", "AD", "BC", "BD", "CD", "L", "T" };
	static const unsigned int name_size = sizeof(name) / sizeof(name[0]);
	const double wscale = 1e+11;

	std::ofstream fstm( file.string().c_str() );
	if( fstm.fail() )
		throw std::runtime_error( "Can't open file " + file.string() );

	fstm.exceptions( std::ifstream::failbit | std::ifstream::badbit );

	fstm << "# Z " << std::setprecision(2) << grid << std::endl;
	fstm << "# W " << std::setprecision(0) << std::scientific << wscale << std::endl;
	if( grid.size() < 1 )
		 return;

	unsigned int result_size = function( grid(0) ).size();

	fstm << "## Alt";
	for( unsigned int i = 0; i < std::min(name_size, result_size); ++i )
		fstm << "\tW(" << name[i] << ")/" <<  std::scientific << wscale;
	fstm << std::endl;

	for( vector_type::size_type i = 0; i < grid.size(); ++i ){
		fstm << std::fixed << std::setprecision(3) << grid(i);
		vector_type result = function( grid(i) );
		fstm << std::scientific << std::setprecision(4);
		for( vector_type::size_type j = 0; j < result.size(); ++j ) {
			fstm << "\t" << result(j) / wscale;
		}
		fstm << std::endl;
	}
}

spectral_factory::spectral_factory(const boost::filesystem::path& wd, const boost::filesystem::path& mass_response, const boost::filesystem::path& ccd_response ):
	m_mass_response(mass_response),
	m_ccd_response(ccd_response),
	m_wdir(wd),
	m_basis_grid( atmos::lg_generator(0.0, 0.04, 0.25), 50 ),
	m_mass_reaction( spectral::load_reaction(mass_response) ),
	m_dimm_reaction( spectral::load_reaction(ccd_response) ) {

	if( ! (boost::filesystem::exists( wd ) && boost::filesystem::is_directory( wd )) )
		throw std::runtime_error( wd.string() + " is not a directory or doesn't exists");
}
spectral_factory::~spectral_factory() {

}
const boost::filesystem::path spectral_factory::get_spectrum_filename( const std::string& spectype ) const {
	std::string spectral_type( spectype );
	boost::to_lower( spectral_type );
	boost::filesystem::path spect( m_wdir / (spectral_type + ".sp") );
	return spect;
}
const spectral_factory::weight_type& spectral_factory::nullexposure_weight( const std::string& spectype, const vector<double>& inner_diam, const vector<double>& outer_diam, double dimmbase, double dimmsize ) {
	storage_type::const_iterator it = m_nullexposure.find( spectype );
	if( it != m_nullexposure.end() )
		return it->second;

	std::pair< vector_type, vector_type > spectral_response = spectral::load_reaction( get_spectrum_filename( spectype ) );

	std::pair< vector_type, vector_type > mass_response = spectral::photon_distribution( spectral_response.first, spectral_response.second, m_mass_reaction.first, m_mass_reaction.second );
	std::pair< vector_type, vector_type > dimm_response = spectral::photon_distribution( spectral_response.first, spectral_response.second, m_dimm_reaction.first, m_dimm_reaction.second );

	std::pair< storage_type::iterator, bool > ret = m_nullexposure.insert( std::make_pair( spectype, weight_type( m_basis_grid, utils::ublas::project( m_basis_grid , atmos::weight::massdimm_weight( mass_response.first, mass_response.second, dimm_response.first, dimm_response.second, inner_diam, outer_diam, dimmsize, dimmbase ) ) ) ) );

	return ret.first->second;
}
const spectral_factory::weight_type& spectral_factory::shortexposure_weight( const std::string& spectype, const vector<double>& inner_diam, const vector<double>& outer_diam ) {
	storage_type::const_iterator it = m_shortexposure.find( spectype );
	if( it != m_shortexposure.end() )
		return it->second;

	std::pair< vector_type, vector_type > spectral_response = spectral::load_reaction( get_spectrum_filename( spectype ) );

	std::pair< vector_type, vector_type > mass_response = spectral::photon_distribution( spectral_response.first, spectral_response.second, m_mass_reaction.first, m_mass_reaction.second );

	std::pair< storage_type::iterator, bool > ret = m_shortexposure.insert( std::make_pair( spectype, weight_type( m_basis_grid, utils::ublas::project( m_basis_grid , atmos::weight::mass_shortexposure_weight( mass_response.first, mass_response.second, inner_diam, outer_diam) ) ) ) );

	return ret.first->second;
}
const spectral_factory::weight_type& spectral_factory::longexposure_weight( const std::string& spectype, const vector<double>& inner_diam, const vector<double>& outer_diam ) {
	storage_type::const_iterator it = m_longexposure.find( spectype );
	if( it != m_longexposure.end() )
		return it->second;

	std::pair< vector_type, vector_type > spectral_response = spectral::load_reaction( get_spectrum_filename( spectype ) );

	std::pair< vector_type, vector_type > mass_response = spectral::photon_distribution( spectral_response.first, spectral_response.second, m_mass_reaction.first, m_mass_reaction.second );

	std::pair< storage_type::iterator, bool > ret = m_longexposure.insert( std::make_pair( spectype, weight_type( m_basis_grid, utils::ublas::project( m_basis_grid , atmos::weight::mass_longexposure_weight( mass_response.first, mass_response.second, inner_diam, outer_diam) ) ) ) );

	return ret.first->second;
}
